
// throughout the algebraic functions below ...

// mul_op will usually be some sort of multiplication. examples:
//
// struct mul_double {
//     double operator() (double const a, double const b) { return a * b; }
// };
//
// struct mul_y_ab {
//     typ_y operator() (typ_a const a, typ_b const b) { return typ_y (a.x * b.x); }
// };

// add_asgn_op will usually be some sort of addition. a void return is also okay,
// but the c++ tradition is to return something from operator+=. examples:
//
// struct add_asgn_double {
//     double operator() (double & z, double const y) { return z += y; }
// };
//
// struct add_asgn_z_zy {
//     typ_z operator() (typ_z & z, typ_y const y) { return typ_z (z.x += y.x); }
// };
//
// add_id { } should be the identity for whatever add_op is
// a likely choice is 0

// legitimate combinations for multiple-type arithmetic as used here:
// type_y  = type_a * type_b;
// type_z += type_y;

// -----------------------------

// outer_product //

template <
    typename mul_oper,
    typename compo_a,
    typename compo_b = compo_a,
    typename compo_y = compo_a,
    typename alloc_a = std::allocator <compo_a>,
    typename alloc_b = std::allocator <compo_b>,
    typename alloc_y = std::allocator <compo_y>
> var_matrix<compo_y,alloc_y> outer_product (
    sub_matrix_base<compo_a,alloc_a> const & mat_a,
    sub_matrix_base<compo_b,alloc_b> const & mat_b
) {
    form const & frm_a {mat_a.get_form()};
    form const & frm_b {mat_b.get_form()};
    form const & frm_y {frm_a.append(frm_b)};
    var_matrix <compo_y,alloc_y> mat_y {frm_y};
    sign_int const count_a {frm_a.count()};
    sign_int const count_b {frm_b.count()};

    mul_oper mul_op;

    for (sign_int n_a {0}; n_a < count_a; ++n_a) {
        compo_a const & k_a {mat_a.get(n_a)};
        for (sign_int n_b {0}; n_b < count_b; ++n_b) {
            compo_b const & k_b {mat_b.get(n_b)};
            compo_y const k_y {mul_op (k_a, k_b)};
            sign_int const n_y {n_a * count_b + n_b};
            mat_y.set (n_y, k_y);
        }
    }
    return mat_y;
}

// contraction //
// unfortunately, we have to establish a vast framework
// allowing us to recourse on the successive subscripts.
// struct contraction_info is a huge parameter block.

template <
    typename add_asgn_oper,
    typename compo_y,
    typename compo_z,
    typename alloc_y,
    typename alloc_z
> struct contraction_info {
    sub_matrix_base<compo_y,alloc_y> const & mat_y;
    var_matrix_base<compo_z,alloc_z> & mat_z;
    vec_bool const & which;
    sign_int const calc_bot; sign_int const calc_top;
    vec_si & vec_y; sign_int & loc_y;
    vec_si & vec_z; sign_int & loc_z;

    contraction_info (
        sub_matrix_base<compo_y,alloc_y> const & par_mat_y,
        var_matrix_base<compo_z,alloc_z> & par_mat_z,
        vec_bool const & par_which,
        sign_int const par_calc_bot, sign_int const par_calc_top,
        vec_si & par_vec_y, sign_int & par_loc_y,
        vec_si & par_vec_z, sign_int & par_loc_z
    ) :
        mat_y {par_mat_y}, mat_z {par_mat_z},
        which {par_which},
        calc_bot {par_calc_bot}, calc_top {par_calc_top},
        vec_y {par_vec_y}, loc_y {par_loc_y},
        vec_z {par_vec_z}, loc_z {par_loc_z}
    { }
};

template <
    typename add_asgn_oper,
    typename compo_y,
    typename compo_z,
    typename alloc_y,
    typename alloc_z
> void contraction_2 (contraction_info
    <add_asgn_oper,compo_y,compo_z,alloc_y,alloc_z> & info)
{
    add_asgn_oper add_asgn_op;

// first pass, mandatory because an additive identity
// is not definable for all possible component types

    sign_int const s0 {info.calc_bot};
    if (check) {
        for (sign_int f {0}; f < info.which.size(); ++f)
            if (info.which.at(f)) info.vec_y.at(f) = s0;
    } else {
        for (sign_int f {0}; f < info.which.size(); ++f)
            if (info.which[f]) info.vec_y[f] = s0;
    }
    compo_z sum {info.mat_y.get(info.vec_y)};

// remaining passes, if any

    for (sign_int s9 {info.calc_bot + 1}; s9 < info.calc_top; ++s9) {
        if (check) {
            for (sign_int f {0}; f < info.which.size(); ++f)
                if (info.which.at(f)) info.vec_y.at(f) = s9;
        } else {
            for (sign_int f {0}; f < info.which.size(); ++f)
                if (info.which[f]) info.vec_y[f] = s9;
        }
        add_asgn_op (sum, info.mat_y.get(info.vec_y));
    }

// passes complete

    info.mat_z.set (info.vec_z, sum);
}

template <
    typename add_asgn_oper,
    typename compo_y,
    typename compo_z,
    typename alloc_y,
    typename alloc_z
> void contraction_1 (contraction_info
    <add_asgn_oper,compo_y,compo_z,alloc_y,alloc_z> & info)
{
    if (info.loc_y == info.which.size())
    { contraction_2 (info); return; }

    if (check) {
        if (info.which.at(info.loc_y)) {
            ++info.loc_y; contraction_1 (info);
            --info.loc_y; return;
        }
    } else {
        if (info.which[info.loc_y]) {
            ++info.loc_y; contraction_1 (info);
            --info.loc_y; return;
        }
    }

    blt const & r {info.mat_y.get_form().get(info.loc_y)};
    sign_int const loop_bot {r.get_bot()};
    sign_int const loop_top {r.get_top()};

    for (sign_int n {loop_bot}; n < loop_top; ++n) {
        if (check) {
            info.vec_y.at(info.loc_y) = n;
            info.vec_z.at(info.loc_z) = n;
        } else {
            info.vec_y[info.loc_y] = n;
            info.vec_z[info.loc_z] = n;
        }
        ++info.loc_y; ++info.loc_z;
        contraction_1 (info);
        --info.loc_z; --info.loc_y;
    }
}

template <
    typename add_asgn_oper,
    typename compo_y,
    typename compo_z = compo_y,
    typename alloc_y = std::allocator <compo_y>,
    typename alloc_z = std::allocator <compo_z>
> var_matrix<compo_z,alloc_z> contraction (
    sub_matrix_base<compo_y,alloc_y> const & mat_y,
    vec_bool const & which
) {
    form const & frm_y {mat_y.get_form()};

    if (check) if (which.size() != frm_y.dimen())
        throw fail {"contraction ... which size does not match dimen"};

    sign_int calc_bot {null_int};
    sign_int calc_top {null_int};
    vector<blt> pre_frm_z;

    for (sign_int n {0}; n < frm_y.dimen(); ++n) {
        blt const & fagn {frm_y.get(n)};
        if (! which.at(n)) {
            pre_frm_z.push_back (fagn);
        } else if (calc_bot == null_int) {
            calc_bot = fagn.get_bot();
            calc_top = fagn.get_top();
        } else if (check) {
            if (fagn.get_bot() != calc_bot)
                throw fail {"contraction ... index calc_bot mismatch"};
            if (fagn.get_top() != calc_top)
                throw fail {"contraction ... index calc_top mismatch"};
        }
    }

    if (calc_bot == null_int) throw fail {"contraction ... no indices"};
    if (calc_top == calc_bot) throw fail {"contraction ... zero length on contraction indices"};

    form const frm_z {pre_frm_z};
    var_matrix <compo_z,alloc_z> mat_z {frm_z};

    vec_si vec_y (frm_y.dimen()); sign_int loc_y {0};
    vec_si vec_z (frm_z.dimen()); sign_int loc_z {0};

    contraction_info<add_asgn_oper,compo_y,compo_z,alloc_y,alloc_z> info {
        mat_y, mat_z, which, calc_bot, calc_top,
        vec_y, loc_y, vec_z, loc_z
    };

    contraction_1 (info);

    return mat_z;

}

// inner_product //
// unfortunately, we have to establish a vast framework
// (even vaster than contraction's framework)
// allowing us to recourse on the successive subscripts.
// struct inner_product_info is a huge parameter block.

template <
    typename mul_oper, typename add_asgn_oper,
    typename compo_a, typename compo_b, typename compo_y, typename compo_z,
    typename alloc_a, typename alloc_b, typename alloc_y, typename alloc_z
> struct inner_product_info {
    sub_matrix_base<compo_a,alloc_a> const & mat_a;
    sub_matrix_base<compo_b,alloc_b> const & mat_b;
    var_matrix_base<compo_z,alloc_z> & mat_z;
    form const & frm_a; form const & frm_b;
    vec_bool const & which; sign_int const calc_bot; sign_int const calc_top;
    vec_si & vec_a; sign_int & loc_a; sign_int const vas;
    vec_si & vec_b; sign_int & loc_b;
    vec_si & vec_z; sign_int & loc_z;

    inner_product_info (
        sub_matrix_base<compo_a,alloc_a> const & par_mat_a,
        sub_matrix_base<compo_b,alloc_b> const & par_mat_b,
        var_matrix_base<compo_z,alloc_z> & par_mat_z,
        vec_bool const & par_which, sign_int const par_calc_bot, sign_int const par_calc_top,
        vec_si & par_vec_a, sign_int & par_loc_a,
        vec_si & par_vec_b, sign_int & par_loc_b,
        vec_si & par_vec_z, sign_int & par_loc_z
    ) :
        mat_a {par_mat_a}, mat_b {par_mat_b}, mat_z {par_mat_z},
        frm_a {mat_a.get_form()}, frm_b {mat_b.get_form()},
        which {par_which}, calc_bot {par_calc_bot}, calc_top {par_calc_top},
        vec_a {par_vec_a}, loc_a {par_loc_a}, vas {sign_int (vec_a.size())},
        vec_b {par_vec_b}, loc_b {par_loc_b},
        vec_z {par_vec_z}, loc_z {par_loc_z}
    { }
};

template <
    typename mul_oper, typename add_asgn_oper,
    typename compo_a, typename compo_b, typename compo_y, typename compo_z,
    typename alloc_a, typename alloc_b, typename alloc_y, typename alloc_z
> void inner_product_3 (inner_product_info <
    mul_oper, add_asgn_oper,
    compo_a, compo_b, compo_y, compo_z,
    alloc_a, alloc_b, alloc_y, alloc_z> & info)
{
    mul_oper mul_op;
    add_asgn_oper add_asgn_op;

// first pass, mandatory because an additive identity
// is not definable for all possible component types

    sign_int const s0 {info.calc_bot};
    if (check) {
        for (sign_int f {0}; f < info.which.size(); ++f) {
            if (info.which.at(f)) {
                if (f < info.vas) info.vec_a.at(f) = s0;
                             else info.vec_b.at(f - info.vas) = s0;
            }
        }
    } else {
        for (sign_int f {0}; f < info.which.size(); ++f) {
            if (info.which[f]) {
                if (f < info.vas) info.vec_a[f] = s0;
                             else info.vec_b[f - info.vas] = s0;
            }
        }
    }
    compo_z sum {
         mul_op (info.mat_a.get(info.vec_a),
                 info.mat_b.get(info.vec_b))
    };

// remaining passes, if any

    for (sign_int s9 {info.calc_bot + 1}; s9 < info.calc_top; ++s9) {
        if (check) {
            for (sign_int f {0}; f < info.which.size(); ++f) {
                if (info.which.at(f)) {
                    if (f < info.vas) info.vec_a.at(f) = s9;
                                 else info.vec_b.at(f - info.vas) = s9;
                }
            }
        } else {
            for (sign_int f {0}; f < info.which.size(); ++f) {
                if (info.which[f]) {
                    if (f < info.vas) info.vec_a[f] = s9;
                                 else info.vec_b[f - info.vas] = s9;
                }
            }
        }
        add_asgn_op (sum,
             mul_op (info.mat_a.get(info.vec_a),
                     info.mat_b.get(info.vec_b)));
    }

// passes complete

    info.mat_z.set (info.vec_z, sum);
}

template <
    typename mul_oper, typename add_asgn_oper,
    typename compo_a, typename compo_b, typename compo_y, typename compo_z,
    typename alloc_a, typename alloc_b, typename alloc_y, typename alloc_z
> void inner_product_2 (inner_product_info<
    mul_oper, add_asgn_oper,
    compo_a, compo_b, compo_y, compo_z,
    alloc_a, alloc_b, alloc_y, alloc_z> & info)
{
    if (info.loc_b == info.frm_b.dimen())
    { inner_product_3 (info); return; }

    if (check) {
        if (info.which.at(info.vas + info.loc_b)) {
            ++info.loc_b; inner_product_2 (info);
            --info.loc_b; return;
        }
    } else {
        if (info.which[info.vas + info.loc_b]) {
            ++info.loc_b; inner_product_2 (info);
            --info.loc_b; return;
        }
    }

    blt const & r {info.frm_b.get(info.loc_b)};
    sign_int const loop_bot {r.get_bot()};
    sign_int const loop_top {r.get_top()};

    for (sign_int n {loop_bot}; n < loop_top; ++n) {
        if (check) {
            info.vec_b.at(info.loc_b) = n;
            info.vec_z.at(info.loc_z) = n;
        } else {
            info.vec_b[info.loc_b] = n;
            info.vec_z[info.loc_z] = n;
        }
        ++info.loc_b; ++info.loc_z;
        inner_product_2 (info);
        --info.loc_z; --info.loc_b;
    }
}

template <
    typename mul_oper, typename add_asgn_oper,
    typename compo_a, typename compo_b, typename compo_y, typename compo_z,
    typename alloc_a, typename alloc_b, typename alloc_y, typename alloc_z
> void inner_product_1 (inner_product_info<
    mul_oper, add_asgn_oper,
    compo_a, compo_b, compo_y, compo_z,
    alloc_a, alloc_b, alloc_y, alloc_z> & info)
{
    if (info.loc_a == info.frm_a.dimen())
    { inner_product_2 (info); return; }

    if (check) {
        if (info.which.at(info.loc_a)) {
            ++info.loc_a; inner_product_1 (info);
            --info.loc_a; return;
        }
    } else {
        if (info.which[info.loc_a]) {
            ++info.loc_a; inner_product_1 (info);
            --info.loc_a; return;
        }
    }

    blt const & r {info.frm_a.get(info.loc_a)};
    sign_int const loop_bot {r.get_bot()};
    sign_int const loop_top {r.get_top()};

    for (sign_int n {loop_bot}; n < loop_top; ++n) {
        if (check) {
            info.vec_a.at(info.loc_a) = n;
            info.vec_z.at(info.loc_z) = n;
        } else {
            info.vec_a[info.loc_a] = n;
            info.vec_z[info.loc_z] = n;
        }
        ++info.loc_a; ++info.loc_z;
        inner_product_1 (info);
        --info.loc_z; --info.loc_a;
    }
}

template <
    typename mul_oper,
    typename add_asgn_oper,
    typename compo_a,
    typename compo_b = compo_a,
    typename compo_y = compo_a,
    typename compo_z = compo_a,
    typename alloc_a = std::allocator <compo_a>,
    typename alloc_b = std::allocator <compo_b>,
    typename alloc_y = std::allocator <compo_y>,
    typename alloc_z = std::allocator <compo_z>
> var_matrix<compo_z,alloc_z> inner_product (
    sub_matrix_base<compo_a,alloc_a> const & mat_a,
    sub_matrix_base<compo_b,alloc_b> const & mat_b,
    vec_bool const & which
) {
    form const & frm_a {mat_a.get_form()};
    form const & frm_b {mat_b.get_form()};
    form const frm_y {frm_a.append(frm_b)};

    if (check) if (which.size() != frm_y.dimen())
        throw fail {"inner_product ... which is wrong size"};

    sign_int calc_bot {null_int};
    sign_int calc_top {null_int};
    vector<blt> pre_frm_z;

    for (sign_int n {0}; n < frm_y.dimen(); ++n) {
        blt const & fygn {frm_y.get(n)};
        if (! which.at(n)) {
            pre_frm_z.push_back (fygn);
        } else if (calc_bot == null_int) {
            calc_bot = fygn.get_bot();
            calc_top = fygn.get_top();
        } else if (check) {
            if (fygn.get_bot() != calc_bot)
                throw fail {"inner_product ... index calc_bot mismatch"};
            if (fygn.get_top() != calc_top)
                throw fail {"inner_product ... index calc_top mismatch"};
        }
    }
    if (calc_bot == null_int) throw fail {"inner_product ... no indices"};
    if (calc_top == calc_bot) throw fail {"inner_product ... zero length on contraction indices"};

    form const frm_z {pre_frm_z};
    var_matrix <compo_z,alloc_z> mat_z {frm_z};

    vec_si vec_a (frm_a.dimen()); sign_int loc_a {0};
    vec_si vec_b (frm_b.dimen()); sign_int loc_b {0};
    vec_si vec_z (frm_z.dimen()); sign_int loc_z {0};

    inner_product_info<
        mul_oper, add_asgn_oper,
        compo_a, compo_b, compo_y, compo_z,
        alloc_a, alloc_b, alloc_y, alloc_z> info
    {
        mat_a, mat_b, mat_z, which, calc_bot, calc_top,
        vec_a, loc_a, vec_b, loc_b, vec_z, loc_z
    };

    inner_product_1 (info);

    return mat_z;
}